/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
dispatcher.cpp   Interface class that hides the dispatcher implementation
                 from the users of the class.
---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "notifier.h"
#include "subscriber.h"
#include "dispatcher.h"
#include "dispprv.h"
#include "subscriber_element.h"
#include "simerr_notify.h"       // error simulation

/********************************
*** Static Data and Functions ***
********************************/

static CDispPrivate *g_pDispPrivate; // there is ONE copy of the dispatcher

/*---------------------------------------------------------------------------
TimerProc      Once the dispatcher is initialized, this proc is called 
               periodically (usually every 50 ms).
---------------------------------------------------------------------------*/
VOID CALLBACK TimerProc(HWND ,         // always NULL in our case
                        UINT ,         // WM_TIMER
                        UINT ,         // timer ID (not relevant in our case)
                        DWORD dwTime)  // system time
{
   static bDispatching = false;
   if (!bDispatching)
   {  // Don't waste time re-entering the OnTimer function
      bDispatching = true;
      try
      {
         if (g_pDispPrivate)
         {
            g_pDispPrivate->OnTimer(dwTime);
            SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION);
         }
      }
      catch(CException *e)
      {
         e->Delete();
      }
      bDispatching = false;
   }
}

/**************************************
*** Construction and Initialization ***
**************************************/

/*---------------------------------------------------------------------------
constructor
---------------------------------------------------------------------------*/
CDispatcher::CDispatcher(void)
{
   g_pDispPrivate = 0;
   m_nTimer = 0;
}

/*---------------------------------------------------------------------------
destructor     Delete any undispatched notifiers, delete lists.
---------------------------------------------------------------------------*/
CDispatcher::~CDispatcher(void)
{
   CSubscriber::SetDispatcher(0);
   CNotifier::SetDispatcher(0);
   if (m_nTimer)
   {
      KillTimer(0, m_nTimer);
   }
   if (g_pDispPrivate)
   {
      delete g_pDispPrivate;
      g_pDispPrivate = 0;
   }
}

/*---------------------------------------------------------------------------
Init           Create a private dispatcher 
---------------------------------------------------------------------------*/
HRESULT CDispatcher::Init(UINT nTimerPeriod)
{
   HRESULT hr = E_OUTOFMEMORY;
   if (g_pDispPrivate != 0)
   {
      hr = E_FAIL;   // only allowed one copy of the dispatcher
   }
   else 
   {
      g_pDispPrivate = new CDispPrivate();
      g_pDispPrivate = SimErrNew(SIMLINE, g_pDispPrivate);
      if (g_pDispPrivate)
      {
         hr = g_pDispPrivate->Init(this);
         hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
         if (SUCCEEDED(hr))
         {
            m_nTimer = ::SetTimer(0, 0, nTimerPeriod, (TIMERPROC) TimerProc);
            m_nTimer = SimErrTimer(SIMLINE, m_nTimer);
            if (m_nTimer == 0)
            {
               hr = ERROR_NO_SYSTEM_RESOURCES;
            }
         }
         if (FAILED(hr))
         {
            delete g_pDispPrivate;
            g_pDispPrivate = 0;
         }
      }
   }
   if (SUCCEEDED(hr))
   {
      CSubscriber::SetDispatcher(this);
      CNotifier::SetDispatcher(this);
   }
   return hr;
}


/*****************************
*** NOTIFIER POST AND SEND ***
*****************************/

/*---------------------------------------------------------------------------
Post           Queue a notifier for subsequent dispatch
---------------------------------------------------------------------------*/
HRESULT CDispatcher::Post(CNotifier *pNotifier)
{
   if (g_pDispPrivate)
   {
      return g_pDispPrivate->Post(pNotifier);
   }
   else
   {
      return E_FAIL;
   }
}
/*---------------------------------------------------------------------------
Send           Send a notifier to all subscribers while caller waits.
---------------------------------------------------------------------------*/
HRESULT CDispatcher::Send(CNotifier *pNotifier, 
                          CNotifier::AUTODELETE bDelete)
{
   if (g_pDispPrivate)
   {
      return g_pDispPrivate->Send(pNotifier, bDelete);
   }
   else
   {
      return E_FAIL;
   }
}

/********************************
*** SUBSCRIBE AND UNSUBSCRIBE ***
********************************/

/*---------------------------------------------------------------------------
IsSubscribed   return true if the subscriber holds a given subscription
---------------------------------------------------------------------------*/
bool CDispatcher::IsSubscribed(CSubscriber *pSubscriber, 
                  UINT nSubscription)
{
   if (g_pDispPrivate)
   {
      return g_pDispPrivate->IsSubscribed(pSubscriber, nSubscription);
   }
   else
   {
      return false;
   }
}
/*---------------------------------------------------------------------------
Subscribe      for a given subscriber, add a subscription to a notifier
---------------------------------------------------------------------------*/
HRESULT CDispatcher::Subscribe(CSubscriber *pSubscriber, 
                               UINT nSubscription)
{
   if (g_pDispPrivate)
   {
      return g_pDispPrivate->Subscribe(pSubscriber, nSubscription);
   }
   else
   {
      return E_FAIL;
   }
}
/*---------------------------------------------------------------------------
Unsubscribe    for a given subscriber, remove a subscription to a notifier
---------------------------------------------------------------------------*/
HRESULT CDispatcher::Unsubscribe(CSubscriber *pSubscriber, 
                                 UINT nSubscription)
{
   if (g_pDispPrivate)
   {
      return g_pDispPrivate->Unsubscribe(pSubscriber, nSubscription);
   }
   else
   {
      return E_FAIL;
   }
}
/*---------------------------------------------------------------------------
UnsubscribeAll for a given subscriber, remove all subscriptions
---------------------------------------------------------------------------*/
HRESULT CDispatcher::UnsubscribeAll(CSubscriber *pSubscriber)
{
   if (g_pDispPrivate)
   {
      return g_pDispPrivate->UnsubscribeAll(pSubscriber);
   }
   else
   {
      return E_FAIL;
   }
}
/*---------------------------------------------------------------------------
SetSubscriberPriority   change the position of a subscriber within the
                        the subscriber list
---------------------------------------------------------------------------*/
HRESULT CDispatcher::SetSubscriberPriority(CSubscriber *pSubscriber,
                              const CSubscriber::SUBSCRIBER_PRIORITY nPriority)
{
   if (g_pDispPrivate)
   {
      return g_pDispPrivate->SetSubscriberPriority(pSubscriber, nPriority);
   }
   else
   {
      return E_FAIL;
   }
}

