/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
dispprv.cpp   CDispPrivate class dispatches CNotifiers to CSubscribers
              This is the workhorse class of the notify DLL. It manages
              lists of subscribers and notifiers, and handles requests
              to add or remove subscriptions, and post notifiers.
---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "notifier.h"
#include "subscriber.h"
#include "dispatcher.h"
#include "subscriber_element.h"
#include "dispprv.h"
#include "simerr_notify.h"       // error simulation

/**************************************
*** Construction and Initialization ***
**************************************/

/*---------------------------------------------------------------------------
constructor
---------------------------------------------------------------------------*/
CDispPrivate::CDispPrivate(void)
{
   m_bInitialized        = false;
   m_bListsInSync        = true;
   m_bDispatching        = false;
   m_pDispatcher         = 0;
   m_pSubscriberMain     = 0;
   m_pSubscriberDispatch = 0;
   m_pNotifierQueue      = 0;
   InitializeCriticalSection(&m_csSubscriber);
   InitializeCriticalSection(&m_csNotifier);
}

/*---------------------------------------------------------------------------
destructor     Delete any undispatched notifiers, delete lists.
---------------------------------------------------------------------------*/
CDispPrivate::~CDispPrivate(void)
{
   if (m_pSubscriberMain)
   {  // delete main subscriber list
      delete m_pSubscriberMain;
   }
   if (m_pSubscriberDispatch)
   {  // delete subscriber list used during dispatch
      delete m_pSubscriberDispatch;
   }
   if (m_pNotifierQueue)
   {  // delete unposted notifiers
	   POSITION pos = m_pNotifierQueue->GetHeadPosition();
	   while (pos)
      {
		   delete m_pNotifierQueue->GetNext(pos);
      }
	   m_pNotifierQueue->RemoveAll();
      delete m_pNotifierQueue;
   }
   DeleteCriticalSection(&m_csNotifier);
   DeleteCriticalSection(&m_csSubscriber);
}

/*---------------------------------------------------------------------------
Init           Allocation of lists, timer creation
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::Init(CDispatcher *pDispatcher)
{
   HRESULT hr = S_OK;

   m_pDispatcher = pDispatcher;
   if (SUCCEEDED(hr))
   {
      m_pSubscriberMain     = new SubscriberList;
      m_pSubscriberDispatch = new SubscriberList;
      m_pNotifierQueue      = new NotifierQueue;
   }
   m_pSubscriberMain     = SimErrNew(SIMLINE, m_pSubscriberMain);
   m_pSubscriberDispatch = SimErrNew(SIMLINE, m_pSubscriberDispatch);
   m_pNotifierQueue      = SimErrNew(SIMLINE, m_pNotifierQueue);
   if ((m_pSubscriberMain==0)     ||
       (m_pSubscriberDispatch==0) ||
       (m_pNotifierQueue==0))
   {
      hr = E_OUTOFMEMORY;
   }
   if (SUCCEEDED(hr))
   {
      m_bInitialized = true;
   }
   return hr;
}

/*****************************
*** NOTIFIER POST AND SEND ***
*****************************/

/*---------------------------------------------------------------------------
Post           Queue a notifier for subsequent dispatch
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::Post(CNotifier *pNotifier)
{
   HRESULT hr = S_OK;
   if (!m_bInitialized)
   {
      hr = E_FAIL;
   }
   else
   {
      hr = InsertNotifier(pNotifier);
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
   }
   return hr;
}
/*---------------------------------------------------------------------------
Send           Send a notifier to all subscribers while caller waits.
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::Send(CNotifier *pNotifier, 
                           CNotifier::AUTODELETE bDelete)
{
   HRESULT hr = S_OK;
   if (!m_bInitialized)
   {
      hr = E_FAIL;
   }
   else
   {
      hr = DispatchToAll(pNotifier);
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
      // delete notifier after sending it, unless instructed not to
      if (bDelete)
      {
         delete pNotifier;
      }
   }
   return hr;
}

/*---------------------------------------------------------------------------
InsertNotifier  Insert a notifier in priority order
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::InsertNotifier(CNotifier *pNew)
{
   HRESULT hr = S_OK;
   EnterCriticalSection(&m_csNotifier);

   // The list is ordered by notifier priority.  Within priority,
   //   later notifiers are posted at the back.  We loop from back
   //   to front and insert the notifier at the end of its priority
   //   queue.
   try
   {
	   POSITION pos = m_pNotifierQueue->GetTailPosition();
      POSITION currPos = pos;
      bool     bInserted = false;
	   while (pos && !bInserted)
	   {
		   CNotifier *pCurr = m_pNotifierQueue->GetPrev(pos);
         // A lower-numbered priority is more important
         // Insert the new notifier after the last notifier
         //   of a lower-numbered priority.
         if (pNew->GetPriority() >= pCurr->GetPriority())
         {
            m_pNotifierQueue->InsertAfter(currPos, pNew);
            SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION);
            bInserted = true;
         }
         currPos = pos;
      }
      if (!bInserted)
      {
         m_pNotifierQueue->AddTail(pNew);
         SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION);
      }
   }
   catch(CException *e)
   {
      e->Delete();
      hr = E_OUTOFMEMORY;
   }
   LeaveCriticalSection(&m_csNotifier);
   return hr;
}

/********************************
*** SUBSCRIBE AND UNSUBSCRIBE ***
********************************/

/*---------------------------------------------------------------------------
IsSubscribed   return true if the subscriber holds a given subscription
---------------------------------------------------------------------------*/
bool CDispPrivate::IsSubscribed(CSubscriber *pSubscriber, 
                               UINT nSubscription)
{
   bool bIsSubscribed = false;
   CSubscriberElement *pElement = 0;
   EnterCriticalSection(&m_csSubscriber);

   // Look for the subscriber in the main list
   POSITION pos = Find(pSubscriber, m_pSubscriberMain);
   if (pos)
   {  // the subscriber is already in the list, only need to update its mask
      pElement = &m_pSubscriberMain->GetAt(pos);
      bIsSubscribed = pElement->m_mask.IsSubscribed(nSubscription);
   }
   LeaveCriticalSection(&m_csSubscriber);
   return bIsSubscribed;
}


/*---------------------------------------------------------------------------
Subscribe  If a subscriber is already subscribed (to another subscription)
           then it will be in the list and only its mask has to be changed.
           If it's not in the list, it has to be added.

           If we're in the middle of a dispatch, we can't add a subscription
           to the dispatcher's private list.  Instead, we modify the main list 
           and set a flag indicating that the two lists differ.  We'll
           update the dispatcher's list after the dispatch completes.
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::Subscribe(CSubscriber *pSubscriber, 
                               UINT nSubscription)
{
   HRESULT hr = S_OK;
   CSubscriberElement *pElement = 0;
   EnterCriticalSection(&m_csSubscriber);

   // Look for the subscriber in the main list
   POSITION pos = Find(pSubscriber, m_pSubscriberMain);

   //-----------------------------------------------------------------------
   // The subscriber is already in the list, only need to update its mask
   //-----------------------------------------------------------------------
   if (pos)
   {  // UPDATE IN MAIN LIST
      pElement = &m_pSubscriberMain->GetAt(pos);
      pElement->m_mask.Subscribe(nSubscription);

      // UPDATE IN DISPATCHER LIST
      // if the two lists are currently in sync and we're not
      //   dispatching, keep them in sync
      if (m_bListsInSync && !m_bDispatching)
      {  
         pos = Find(pSubscriber, m_pSubscriberDispatch);
         ASSERT(pos != 0);  // lists are in sync, should be able find it
         pElement = &m_pSubscriberDispatch->GetAt(pos);
         pElement->m_mask.Subscribe(nSubscription);
      }
      // if we were dispatching, we couldn't keep the lists in sync
      else
      {
         m_bListsInSync = false;
      }
   }
   //-----------------------------------------------------------------------
   // The subscriber is new, add it to the subscriber list
   //-----------------------------------------------------------------------
   else
   {  // ADD TO MAIN LIST
      CSubscriberElement newElement(pSubscriber);
      newElement.m_mask.Subscribe(nSubscription);
      newElement.m_nPriority = pSubscriber->GetSubscriberPriority();
      hr = InsertSubscriber(newElement, m_pSubscriberMain);
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
      if (SUCCEEDED(hr))
      {
         // ADD TO DISPATCHER LIST
         // if the two lists are currently in sync and we're not
         //   dispatching, keep them in sync
         if (m_bListsInSync && !m_bDispatching)
         {  
            hr = InsertSubscriber(newElement, m_pSubscriberDispatch);
            hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
            if (FAILED(hr))
            {
               m_bListsInSync = false;
            }
         }
         // if we were dispatching, we couldn't keep the lists in sync
         else
         {
            m_bListsInSync = false;
         }
      }
   }
   if (FAILED(hr))
   {
      m_bListsInSync = false;
   }
   ASSERT_IN_SYNC();
   LeaveCriticalSection(&m_csSubscriber);
   return hr;
}

/*---------------------------------------------------------------------------
Unsubscribe  Update the subscription mask for the subscriber.  If the mask
             becomes zero, i.e. we removed the last subscription, remove the
             subscriber record completely.
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::Unsubscribe(CSubscriber *pSubscriber, 
                                 UINT nSubscription)
{
   HRESULT  hr = S_OK;
   CSubscriberElement *pElement = 0;
   EnterCriticalSection(&m_csSubscriber);

   // Find the subscriber, clear the bit in its mask.
   // If the mask is now zero, remove the subscriber from the list.
   POSITION pos = Find(pSubscriber, m_pSubscriberMain);
   if (pos)
   {
      // Clear the subscription bit
      pElement = &m_pSubscriberMain->GetAt(pos);
      pElement->m_mask.Unsubscribe(nSubscription);
      // If the subscriber has no more subscriptions, remove it
      if (pElement->IsMaskClear())
      {
         m_pSubscriberMain->RemoveAt(pos);
      }
   }
   // We keep the mask in the dispatch subscriber list up-to-date 
   // This is important because the moment a subscriber unsubscribes they
   //   must no longer receive notifiers.  For one thing, they may be 
   //   unsubscribing because they're being deleted!
   // But we can't remove subscriber records if a dispatch is active because
   //   that can foul up the iteration through the list.  
   pos = Find(pSubscriber, m_pSubscriberDispatch);
   if (pos)
   {  // found the subscriber, clear the bit in its subscription mask
      pElement = &m_pSubscriberDispatch->GetAt(pos);
      pElement->m_mask.Unsubscribe(nSubscription);
      // if the mask is now zero, remove the subscriber
      if (pElement->IsMaskClear())
      {
         if (!m_bDispatching && m_bListsInSync)
         {  // if not dispatching, remove from the dispatcher list
            m_pSubscriberDispatch->RemoveAt(pos);
         }
         else
         {
            m_bListsInSync = false;
         }
      }
   }
   if (FAILED(hr))
   {
      m_bListsInSync = false;
   }
   ASSERT_IN_SYNC();
   LeaveCriticalSection(&m_csSubscriber);
   return hr;
}
/*---------------------------------------------------------------------------
UnsubscribeAll    Remove a subscriber from the subscriber list.  This removes
                  all their subscriptions.
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::UnsubscribeAll(CSubscriber *pSubscriber)
{
   HRESULT  hr = S_OK;
   CSubscriberElement *pElement = 0;
   EnterCriticalSection(&m_csSubscriber);

   // Find the subscriber and remove it from the list
   POSITION pos = Find(pSubscriber, m_pSubscriberMain);
   if (pos)
   {
      m_pSubscriberMain->RemoveAt(pos);
   }
   // We clear the mask in the dispatch subscriber list.
   // This is important because the moment a subscriber unsubscribes they
   //   must no longer receive notifiers.  For one thing, they be unsubscribing
   //   because they're being deleted!
   // But we can't remove subscribers if a dispatch is active because
   //   that can foul up the iteration through the list.  
   pos = Find(pSubscriber, m_pSubscriberDispatch);
   if (pos)
   {  // found the subscriber, clear its subscription mask
      pElement = &m_pSubscriberDispatch->GetAt(pos);
      pElement->m_mask.Clear();
      if (!m_bDispatching && m_bListsInSync)
      {  // if not dispatching, remove from the dispatcher list
         m_pSubscriberDispatch->RemoveAt(pos);
      }
      else
      {
         m_bListsInSync = false;
      }
   }
   if (FAILED(hr))
   {
      m_bListsInSync = false;
   }
   ASSERT_IN_SYNC();
   LeaveCriticalSection(&m_csSubscriber);
   return hr;
}
/*---------------------------------------------------------------------------
SetSubscriberPriority     change subscriber priority by re-inserting in list
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::SetSubscriberPriority(CSubscriber *pSubscriber,
                     const CSubscriber::SUBSCRIBER_PRIORITY nPriority)
{
   HRESULT  hr = S_OK;
   CSubscriberElement *pElement = 0;
   EnterCriticalSection(&m_csSubscriber);

   POSITION pos = Find(pSubscriber, m_pSubscriberMain);
   if (pos)
   {  // make a copy of the element we found
      CSubscriberElement element = m_pSubscriberMain->GetAt(pos);
      if (element.m_nPriority != nPriority)
      {  // SET PRIORITY IN MAIN LIST
         // remove and re-insert the element
         m_pSubscriberMain->RemoveAt(pos);
         hr = InsertSubscriber(element, m_pSubscriberMain);
         hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
         if (SUCCEEDED(hr))
         {
            // SET PRIORITY IN DISPATCHER LIST
            // if the two lists are currently in sync and we're not
            //   dispatching, keep them in sync
            if (m_bListsInSync && !m_bDispatching)
            {  // lists are in sync, we had better find this subscriber!!
               pos = Find(pSubscriber, m_pSubscriberDispatch);
               ASSERT(pos); // error: lists were not really in sync
               if (pos)
               {  // remove and re-insert in the dispatch list
                  m_pSubscriberDispatch->RemoveAt(pos);
                  hr = InsertSubscriber(element, m_pSubscriberDispatch);
                  hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
                  if (FAILED(hr))
                  {
                     m_bListsInSync = false;
                  }
               }
            }
         }
      }
   }
   if (FAILED(hr))
   {
      m_bListsInSync = false;
   }
   ASSERT_IN_SYNC();
   LeaveCriticalSection(&m_csSubscriber);
   return hr;
}

/*---------------------------------------------------------------------------
Find           Find a subscriber in the subscriber list.
---------------------------------------------------------------------------*/
POSITION CDispPrivate::Find(const CSubscriber *pSubscriber, 
                           const SubscriberList *pList)
{
   POSITION pos = pList->GetHeadPosition();
   POSITION currPos = pos;
	while (pos)
	{
		const CSubscriberElement &element = pList->GetNext(pos);
      if (element.m_pSubscriber == pSubscriber)
      {
         break;
      }
      currPos = pos;
   }
   return currPos;
}

/*---------------------------------------------------------------------------
InsertSubscriber  Insert a subscriber in priority order
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::InsertSubscriber(CSubscriberElement &newElement,
                                       SubscriberList *pList)
{
   HRESULT hr = S_OK;
   CSubscriberElement *pCurrElement = 0;

   // The list is ordered by subscriber priority.  Within priority,
   //   later subscriber are posted at the back.  We loop from back
   //   to front and insert the subscriber at the end of its priority
   //   queue.
   try
   {
	   POSITION pos = pList->GetTailPosition();
      POSITION currPos = pos;
      bool     bInserted = false;
	   while (pos && !bInserted)
	   {
		   pCurrElement = &pList->GetPrev(pos);
         // if the subscriber mask is zero, the subscriber pointer isn't valid
         if (!pCurrElement->IsMaskClear())
         {
            // A lower-numbered priority is more important
            // Insert the new notifier after the last notifier
            //   that is either more important or earlier
            if (newElement.m_nPriority >= pCurrElement->m_nPriority)
            {
               pList->InsertAfter(currPos, newElement);
               SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION);
               bInserted = true;
            }
         }
         currPos = pos;
      }
      if (!bInserted)
      {
         pList->AddTail(newElement);
         SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION);
      }
   }
   catch(CException *e)
   {
      e->Delete();
      hr = E_OUTOFMEMORY;
   }
   return hr;
}

/******************
*** DISPATCHING ***
******************/

/*---------------------------------------------------------------------------
OnTimer        Dispatch all notifiers to their subscribers.  
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::OnTimer(DWORD )
{
   HRESULT hr = S_OK;
   if (!m_bInitialized)
   {
      hr = E_FAIL;
   }
   else 
   {
      // Dispatch all pending notifiers
      while (SUCCEEDED(hr) && !m_pNotifierQueue->IsEmpty())
      {  // lock notifiers, remove top notifier and dispatch to subscribers
         EnterCriticalSection(&m_csNotifier);
         CNotifier *pNotifier = m_pNotifierQueue->RemoveHead();
         LeaveCriticalSection(&m_csNotifier);

         hr = DispatchToAll(pNotifier);
         hr = SimErrResult(SIMLINE, hr, E_FAIL);
         delete pNotifier;
      }
   }
   return hr;
}

/*---------------------------------------------------------------------------
DispatchToAll  Dispatch a notifier to all subscribers.
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::DispatchToAll(CNotifier *pNotifier)
{
   HRESULT hr = S_OK;
   CSubscriberElement *pElement = 0;

   // Prevent other threads from updating the main subscriber list or from
   //   deleting a subscriber during a dispatch.
   // This thread may still update the main list, but may not insert
   //   or delete records from the dispatcher's list.
   EnterCriticalSection(&m_csSubscriber);

   bool bWasDispatching = m_bDispatching;
   m_bDispatching = true;  // stops this thread from updating dispatcher list
   try
   {
      // The dispatcher has its own copy of the subscriber list.  Before a 
      //   dispatch, check if the main list changed and if so, update the 
      //   dispatcher's list.
      hr = UpdateDispatcherList(); 
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);

      // Dispatch the notifier to each subscriber
      if (SUCCEEDED(hr))
      {
         POSITION pos = m_pSubscriberDispatch->GetHeadPosition();
	      while (pos)
	      {
		      pElement = &m_pSubscriberDispatch->GetNext(pos);
            if (pElement->IsSubscribed(pNotifier->GetClass()))
            {
               hr = m_pDispatcher->DispatchToOne(pNotifier, pElement->m_pSubscriber);
               hr = SimErrResult(SIMLINE, hr, E_FAIL);
               SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION);
            }
         }
      }
   }
   catch(CException *e)
   {
      e->Delete();
      hr = E_FAIL;
   }
   m_bDispatching = bWasDispatching;
   LeaveCriticalSection(&m_csSubscriber);
   return hr;
}

/*---------------------------------------------------------------------------
UpdateDispatcherList  If the subscriber queue has changed since the last 
                      update, copy it to the dispatcher's private list.
---------------------------------------------------------------------------*/
HRESULT CDispPrivate::UpdateDispatcherList(void)
{
   HRESULT hr = S_OK;
   CSubscriberElement *pElement = 0;
   try
   {
      if (!m_bListsInSync)
      {  // If the list is out of sync, erase it and copy the main list.
         m_pSubscriberDispatch->RemoveAll();
         POSITION pos = m_pSubscriberMain->GetHeadPosition();
	      while (pos)
	      {
		      pElement = &m_pSubscriberMain->GetNext(pos);
            m_pSubscriberDispatch->AddTail(*pElement);
            SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION);
         }
         m_bListsInSync = true;
      }
   }
   catch(CException *e)
   {
      e->Delete();
      hr = E_OUTOFMEMORY;
   }
   if (FAILED(hr))
   {
      m_bListsInSync = false;
   }
   ASSERT_IN_SYNC();
   return hr;
}

/*---------------------------------------------------------------------------
DumpSubscriberList Log the contents of a subscriber list. 
                   Dump functions require that ERRLOG is #defined and the
                   utilities DLL is linked, so debug data can be written to
                   the circular log file.
---------------------------------------------------------------------------*/
#if defined(_DEBUG) && defined(ERRLOG)
void CDispPrivate::DumpSubscriberList(char *msg, SubscriberList *pList) 
{
   static char *g_listName[] = { "main", "disp" };
   int nListIndex = (pList==m_pSubscriberMain) ? 0 : 1;
   LOGPRINT(("Dump %s Subscriber List %s %d entries", 
      g_listName[nListIndex],
      msg, pList->GetCount()));

   POSITION pos = pList->GetHeadPosition();
   char strMask[80];
   int i=0;
	while (pos)
	{
		const CSubscriberElement &element = pList->GetNext(pos);
      LOGPRINT(("%2d: %2svalid, pri=%d, mask=%s",
         ++i, 
         (element.IsMaskClear()) ? "in" : "", 
          element.m_nPriority,
          element.DumpMask(strMask,80)));
   }
}
#endif

/*---------------------------------------------------------------------------
AssertInSync Validate that the two lists are in sync
---------------------------------------------------------------------------*/
#ifdef _DEBUG
void CDispPrivate::AssertInSync(void) 
{
   if (m_bListsInSync)
   {  // only assert if they're supposed to be in sync
      POSITION posMain = m_pSubscriberMain->GetHeadPosition();
      POSITION posDisp = m_pSubscriberDispatch->GetHeadPosition();
	   while (posMain && posDisp)
	   {
         CSubscriberElement &eltMain = m_pSubscriberMain->GetNext(posMain);
         CSubscriberElement &eltDisp = m_pSubscriberDispatch->GetNext(posDisp);
         if (eltMain.m_pSubscriber != eltDisp.m_pSubscriber)
         {
            ASSERT(false);
         }
         if (eltMain.m_mask != eltDisp.m_mask)
         {
            ASSERT(false);
         }
      }
      if (posMain || posDisp)
      {
         ASSERT(false);
      }
   }
}
#endif


