/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
dispprv.h   CDispPrivate class definition
            This is the private interface for the dispatcher.  
            By using a public class, CDispatcher, which delegates to this
            private class, it reduces the size and complexity of the header
            files exposed to client programs.
---------------------------------------------------------------------------*/
#ifndef DISPPRV_H
#define DISPPRV_H

#include "notifier.h"
#include "subscriber.h"

// Forward declarations and typedefs
class CSubscriberElement;
class CDispatcher;

typedef CList<CSubscriberElement, CSubscriberElement &> SubscriberList;
typedef CList<CNotifier *, CNotifier *> NotifierQueue;

//---------------------------------------------------------------------------
class CDispPrivate
{
public:        // Constructors and destructors
   CDispPrivate(void);
   virtual ~CDispPrivate(void);
   HRESULT Init(CDispatcher *pDispatcher); 

   // Notifier post and send
      // Post: queue a notifier for subsequent dispatch
   HRESULT Post(CNotifier *pNotifier); 
      // Send: immediately dispatch a notifier
   HRESULT Send(CNotifier *pNotifier, 
                CNotifier::AUTODELETE bDelete);

   // Subscribe and Unsubscribe
      // IsSubscribed: return true if a subscriber will be sent this notifier 
   bool IsSubscribed(CSubscriber *pSubscriber, 
                     UINT nSubscription);
      // Subscribe: add subscriber to the list that receives this notifier
   HRESULT Subscribe(CSubscriber *pSubscriber, 
                     UINT nSubscription);
      // Unsubscribe: no longer send this notifier type to the subscriber
   HRESULT Unsubscribe(CSubscriber *pSubscriber, 
                       UINT nSubscription);
      // UnsubscribeAll: send no notifiers to this subscriber
   HRESULT UnsubscribeAll(CSubscriber *pSubscriber);
      // SetSubscriberPriority:  move this subscriber ahead or behind other
      //                         subscribers in the list
   HRESULT SetSubscriberPriority(CSubscriber *pSubscriber,
                                 const CSubscriber::SUBSCRIBER_PRIORITY nPriority);
      // OnTimer: called by TimerProc to dispatch notifiers
   HRESULT OnTimer(DWORD dwTime);
protected:     // Implementation
   // Notifier post and send
      // InsertNotifier: add notifier to the notifier list
   HRESULT InsertNotifier(CNotifier *pNewNotifier);

   // Subscribe and Unsubscribe
      // Find: find a subscriber in a given subscriber list
   POSITION Find(const CSubscriber *pSubscriber, 
                 const SubscriberList *pList);
      // InsertSubscriber: add a subscriber to a given subscriber list
   HRESULT InsertSubscriber(CSubscriberElement &newElement,
                            SubscriberList *pList);

   // Dispatching functions
      // DispatchToAll: dispatch one notifier to all appropriate subscribers
   HRESULT DispatchToAll(CNotifier *pNotifier);
      // UpdateDispatcherList: match the dispatcher's subscriber list to the
      //                       main subscriber list
   HRESULT UpdateDispatcherList(void);

private:     // Data members
   bool              m_bInitialized;   // true if init was successful
   CDispatcher      *m_pDispatcher;    // the interface class that delegates to us

   // m_SubscriberMain is a sorted list of subscribers that is always accurate
   SubscriberList   *m_pSubscriberMain;   
   // m_pSubscriberDispatch is a sorted list of subscribers just for dispatching.
   SubscriberList   *m_pSubscriberDispatch;
   bool              m_bListsInSync;   // true if main list matches dispatch list
   bool              m_bDispatching;   // true if currently dispatching notifiers
   CRITICAL_SECTION  m_csSubscriber;   // protects both subscriber lists during updates 
                                       //   and during dispatching

   // m_pNotifierQueue holds pending notifiers sorted by priority and order of posting
   NotifierQueue    *m_pNotifierQueue;
   CRITICAL_SECTION  m_csNotifier;     // protects the notifier queue

#ifdef _DEBUG  
   // verify the main and dispatcher lists are in sync, if they're supposed to be
   void    AssertInSync(void);
   // dump the contents of a subscriber list to the log
   void    DumpSubscriberList(char *msg, SubscriberList *pList);
#endif

};  // end of CDispPrivate

#ifdef _DEBUG
#define ASSERT_IN_SYNC()  AssertInSync()
#define DUMP_SUBSCRIBER_LIST(msg, plist)  DumpSubscriberList(msg, plist)
#else
#define ASSERT_IN_SYNC()
#define DUMP_SUBSCRIBER_LIST(msg, plist)
#endif //_DEBUG

#endif // DISPPRV_H
