/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
graph.cpp   CGraph class implementation
 
Draw a line-graph that displays temperature changes over time.
---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "thermdef.h"
#include "graph.h"
#include "temperatureNotifier.h"
#include "unitsNotifier.h"
#include "setpointNotifier.h"
#include "requestNotifier.h"
#include "scale.h"
#include "simerr_notify.h"       // error simulation

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*---------------------------------------------------------------------------
Message Map
---------------------------------------------------------------------------*/
BEGIN_MESSAGE_MAP(CGraph, CWnd)
	//{{AFX_MSG_MAP(CGraph)
	ON_WM_PAINT()
	ON_WM_TIMER()
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*---------------------------------------------------------------------------
CGraph   constructor
---------------------------------------------------------------------------*/
CGraph::CGraph(void)
       :CWnd(), CThermSubscriber()
{
   m_fCurrTemp        = 34.0;
   m_nTimer           = 0;
   m_nTempCount       = 0;
   m_dwNextTime       = 0;
   m_dwUpdateInterval = 100;  // in milliseconds
   m_rectWindow.SetRectEmpty();
   m_fTempArray.SetSize(0);
   m_fSPArray.SetSize(0);
   m_units = CELSIUS;  
   m_fSetPoint = ToCelsius(75.0);
   m_pScale = GetAppScale();
}
/*---------------------------------------------------------------------------
CGraph   destructor
---------------------------------------------------------------------------*/
CGraph::~CGraph(void)
{
}
/*---------------------------------------------------------------------------
Init           create timer for graph update
               create and size the window
               create pens & brushes for drawing
---------------------------------------------------------------------------*/
HRESULT CGraph::Init(CWnd *pParent, UINT nID)
{
   HRESULT hr = S_OK;
   BOOL bCreated = FALSE;
   
	m_penWhite.CreateStockObject(WHITE_PEN);
	m_penBlack.CreateStockObject(BLACK_PEN);
	m_penLtGray.CreatePen(PS_SOLID, 1, RGB_LIGHT_GRAY);
	m_penDkGray.CreatePen(PS_SOLID, 1, RGB_DARK_GRAY);
   m_penLtBlue.CreatePen(PS_SOLID, 1, RGB_GRIDLINE_COLOR);
   m_penMercury.CreatePen(PS_SOLID, 2, RGB_MERCURY_COLOR);
   m_brushLtGray.CreateStockObject(LTGRAY_BRUSH);

   // create the window
   bCreated = CreateEx(WS_EX_STATICEDGE,
                       NULL, "", 
                       WS_CHILD | WS_VISIBLE,
                       m_rectWindow, pParent, nID,
                       0);
   bCreated = SimErrCreateWin(SIMLINE, bCreated, this);
   if (!bCreated)
   {
      hr = E_OUTOFMEMORY;
   }
   if (SUCCEEDED(hr))
   {  // Setup a timer that triggers ten times a second
      m_nTimer = SetTimer(1000,  // timer ID
                          100,   // timer interval in milliseconds
                          0);    // use WM_TIMER instead of a proc
      m_nTimer = SimErrTimer(SIMLINE, m_nTimer);
      if (m_nTimer == 0)
      {
         hr = ERROR_NO_SYSTEM_RESOURCES;
      }
   }
   if (SUCCEEDED(hr))
   {  // ask for the temperature, units, and current setpoint
      Subscribe(TEMPERATURE_NOTIFIER);
      Subscribe(UNITS_NOTIFIER);
      Subscribe(SETPOINT_NOTIFIER);
   }

   return hr;
}
/*---------------------------------------------------------------------------
OnDestroy      kill the timer before the window is destroyed
---------------------------------------------------------------------------*/
void CGraph::OnDestroy() 
{
   if (m_nTimer)
   {
      KillTimer(m_nTimer);
   }
	CWnd::OnDestroy();
}

/*---------------------------------------------------------------------------
OnPaint        WM_PAINT - draw the graph
               draw a gray background then draw the grid lines
---------------------------------------------------------------------------*/
void CGraph::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
   CRect rectIntersect;
   CRect rectInvalid;

   rectInvalid = dc.m_ps.rcPaint;
   dc.FillRect(&rectInvalid, &m_brushLtGray);

   DrawGridLines(&dc, rectInvalid);
   DrawTemperature(&dc, rectInvalid);
#ifdef RELEASE3
   DrawSetPoint(&dc, rectInvalid);
#endif
}

/*---------------------------------------------------------------------------
DrawGridLines  Draw the blue lines of the grid
---------------------------------------------------------------------------*/
void CGraph::DrawGridLines(CDC *pDC, CRect &rectInvalid)
{
   CRect clientRect;
   GetClientRect(&clientRect);

   int nInset = m_pScale->GetInset(clientRect.Height());
   clientRect.top    += nInset;
   clientRect.bottom -= nInset;

   int width = clientRect.Width();
   int interval = m_pScale->GetCellSize().cy;
   int halfInterval = interval / 2;

	CPen *pOldPen = (CPen *) pDC->SelectObject(&m_penLtBlue);
   for (int y = clientRect.top + halfInterval;
        y < clientRect.bottom - halfInterval;
        y += interval)
   {
      pDC->MoveTo(rectInvalid.left,   y);
      pDC->LineTo(rectInvalid.right,  y);
   }
	pDC->SelectObject(pOldPen);
}

/*---------------------------------------------------------------------------
DrawTemperature   Draw the temperature data points
---------------------------------------------------------------------------*/
void CGraph::DrawTemperature(CDC *pDC, CRect &rectInvalid)
{
   CRect clientRect;
   int   i;       // data array index; element zero is most recent
   int   x, y;    // position of datapoint on the graph
	CPen *pOldPen = (CPen *) pDC->SelectObject(&m_penMercury);

   GetClientRect(&clientRect);
   int top = m_pScale->GetInset(clientRect.Height());

   if (m_nTempCount)
   {
      i = 0;
      x = clientRect.right;
      y = top + m_pScale->CelsiusToPixels(m_fTempArray[i]);
      pDC->MoveTo(x, y);

      if (rectInvalid.left > clientRect.left)
      {
         rectInvalid.left--;
      }
      while ((--x > rectInvalid.left) &&
             (++i < m_nTempCount))
      {
         y = top + m_pScale->CelsiusToPixels(m_fTempArray[i]);
         pDC->LineTo(x, y);
      }
   }
	pDC->SelectObject(pOldPen);
}

/*---------------------------------------------------------------------------
DrawSetPoint   Draw the set points
---------------------------------------------------------------------------*/
void CGraph::DrawSetPoint(CDC *pDC, CRect &rectInvalid)
{
   CRect clientRect;
   int   i;       // data array index; element zero is most recent
   int   x, y;    // position of datapoint on the graph
   int   x0,y0;   // previous datapoint position
	CPen *pOldPen = (CPen *) pDC->SelectObject(&m_penWhite);

   GetClientRect(&clientRect);
   int top = m_pScale->GetInset(clientRect.Height());

   if (m_nTempCount)
   {
      i = 0;
      x0 = x = clientRect.right;
      y0 = y = top + m_pScale->CelsiusToPixels(m_fSPArray[i]);

      if (rectInvalid.left > clientRect.left)
      {
         rectInvalid.left--;
      }
      while ((--x > rectInvalid.left) &&
             (++i < m_nTempCount))
      {
         y = top + m_pScale->CelsiusToPixels(m_fSPArray[i]);

         // A 3D line is white/ltgray/gray/black
         pDC->SelectObject(&m_penWhite);     // white above
         pDC->MoveTo(x0, y0-2);
         pDC->LineTo(x,  y-2);

         pDC->SelectObject(&m_penLtGray);     // 2-pixel lt gray
         pDC->MoveTo(x0, y0-1);
         pDC->LineTo(x,  y-1);

         pDC->SelectObject(&m_penLtGray);     
         pDC->MoveTo(x0, y0);
         pDC->LineTo(x,  y);

         pDC->SelectObject(&m_penDkGray);     // dk gray
         pDC->MoveTo(x0, y0+1);
         pDC->LineTo(x,  y+1);

         pDC->SelectObject(&m_penBlack);     // black below
         pDC->MoveTo(x0, y0+2);
         pDC->LineTo(x,  y+2);

         x0 = x;  // save current 
         y0 = y;  //   end of line
      }
   }
	pDC->SelectObject(pOldPen);
}


/*---------------------------------------------------------------------------
SetRect        Set the window rectangle
---------------------------------------------------------------------------*/
void CGraph::SetRect(CRect &rect)
{
   if (GetSafeHwnd() && (m_rectWindow != rect))
   {
      MoveWindow(rect, TRUE);

      // We store one temperature for each pixel in the graph
      //   so we resize the temperature data array when the 
      //   window is resized.
      try
      {  // altering the array size can throw a memory exception
         CRect clientRect;
         int   newSize;

         GetClientRect(&clientRect);
         newSize = clientRect.Width();
         if (newSize != m_fTempArray.GetSize())
         {  // resize the array
            m_fTempArray.SetSize(newSize);
            SimErrException(SIMLINE, SIMERR_MEMORY_EXCEPTION); 
            
            if (m_nTempCount > newSize)
            {  // if we made the array smaller, 
               //   we may have thrown out some data points
               m_nTempCount = newSize;
            }
         }
         if (newSize != m_fSPArray.GetSize())
         {  // resize the array
            m_fSPArray.SetSize(newSize);
         }
      }
      catch(CException *e)
      {  // if we can't resize the array, we just won't show the data points
         // a real application would report the error.
         e->Delete();
      }
   }
   m_rectWindow = rect;

}

/*---------------------------------------------------------------------------
GetDesiredSize increase given width and height as necessary to satisfy 
               minimum size requirements
---------------------------------------------------------------------------*/
void CGraph::GetDesiredSize(int &, int &height)
{
   height = max(height, MIN_METER_HEIGHT);  
}

/*---------------------------------------------------------------------------
OnTemperatureNotifier   
---------------------------------------------------------------------------*/
void CGraph::OnTemperatureNotifier(CTemperatureNotifier *pNotifier)
{
   m_fCurrTemp = pNotifier->m_celsius;
}
/*---------------------------------------------------------------------------
OnSetPointNotifier   
---------------------------------------------------------------------------*/
void CGraph::OnSetPointNotifier(CSetPointNotifier *pNotifier)
{
   m_fSetPoint = pNotifier->m_celsius;
}


/*---------------------------------------------------------------------------
OnTimer        Check the temperature periodically and update the graph.
               This function may trigger faster than the update interval.
               The m_fTempArray acts like a ring buffer.  Periodically we
               throw out the oldest data point and insert a new one.
---------------------------------------------------------------------------*/
void CGraph::OnTimer(UINT nIDEvent) 
{
   static int nCount = 0;
   if (GetTickCount() > m_dwNextTime)
   {
      int size = m_fTempArray.GetSize();
      ASSERT(size == m_fSPArray.GetSize());
      if (size > 0)
      {
         if (m_nTempCount >= size)
         {  // if we've filled the array, pop the oldest entry
            m_fTempArray.RemoveAt(--m_nTempCount);
            m_fSPArray.RemoveAt(--m_nTempCount);
         }
         // add the most recent entry to the beginning
         ASSERT(m_nTempCount < size);
	      m_fTempArray.InsertAt(0, m_fCurrTemp);
	      m_fSPArray.InsertAt(0, m_fSetPoint);
         ++m_nTempCount;

         // redraw every n pixels
         if (++nCount >= 1)
         {
            CRect client;
            ScrollWindow(-1, 0);
            GetClientRect(&client);
            client.left = client.right - 1;
            InvalidateRect(client, FALSE);
            UpdateWindow();
            nCount = 0;
         }
      }
      m_dwNextTime = GetTickCount() + m_dwUpdateInterval;
   }

	CWnd::OnTimer(nIDEvent);
}

/*---------------------------------------------------------------------------
OnUnitsNotifier   change from celsius to fahrenheit
---------------------------------------------------------------------------*/
void CGraph::OnUnitsNotifier(CUnitsNotifier *pNotifier)
{
   if (m_units != pNotifier->m_units)
   {
      // change the units
      m_units = pNotifier->m_units;

      Invalidate(FALSE);
   }
}

