/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
meter.cpp   CThermometer class implementation

This class draws a liquid-mercury thermometer graphically representing
the current temperature.  It is display only and does not respond to 
user input.
---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "thermdef.h"
#include "meter.h"
#include "temperatureNotifier.h"
#include "unitsNotifier.h"
#include "requestNotifier.h"
#include "scale.h"
#include "simerr_notify.h"       // error simulation

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*---------------------------------------------------------------------------
Message Map
---------------------------------------------------------------------------*/
BEGIN_MESSAGE_MAP(CThermometer, CWnd)
	//{{AFX_MSG_MAP(CThermometer)
	ON_WM_PAINT()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*---------------------------------------------------------------------------
CThermometer   constructor
---------------------------------------------------------------------------*/
CThermometer::CThermometer(void)
             :CWnd(), CThermSubscriber()
{
   m_rectWindow.SetRectEmpty();
   m_rectMercury.SetRectEmpty();

   m_fCurrTemp  = 75.0;
   m_units      = CELSIUS;   
   m_gradeLong  = 0;
   m_gradeShort = 0;
   m_pScale     = GetAppScale();
}
/*---------------------------------------------------------------------------
CThermometer   destructor
---------------------------------------------------------------------------*/
CThermometer::~CThermometer(void)
{
}
/*---------------------------------------------------------------------------
Init           create and size the window and 
               create pens & brushes for drawing
---------------------------------------------------------------------------*/
HRESULT CThermometer::Init(CWnd *pParent, UINT nID)
{
   HRESULT hr = S_OK;
   BOOL bCreated = FALSE;
   
   // Create pens, brushes, and fonts
	m_penBlack.CreateStockObject(BLACK_PEN);
   m_penDkGray.CreatePen(PS_SOLID, 1, RGB_DARK_GRAY);
   m_penLtBlue.CreatePen(PS_SOLID, 1, RGB_GRIDLINE_COLOR);
   m_brushLtGray.CreateStockObject(LTGRAY_BRUSH);
   m_brushMercury.CreateSolidBrush(RGB_MERCURY_COLOR);
   m_font.CreateStockObject(ANSI_VAR_FONT);

   // create the window
   bCreated = Create(NULL, "", 
                     WS_CHILD | WS_VISIBLE,
                     m_rectWindow, pParent, nID,
                     0);
   bCreated = SimErrCreateWin(SIMLINE, bCreated, this);
   if (!bCreated)
   {
      hr = E_OUTOFMEMORY;
   }

   // find out the temperature now, and when it changes
   if (SUCCEEDED(hr))
   {
      Subscribe(UNITS_NOTIFIER);
      Subscribe(TEMPERATURE_NOTIFIER);
      RequestNotifierPost(TEMPERATURE_NOTIFIER);
      RequestNotifierPost(UNITS_NOTIFIER);
   }

   return hr;
}

/*---------------------------------------------------------------------------
Paint          Fill the background with gray, then draw the bar of mercury,
               the blue gradation lines on either side of it, and the 
               numeric scale to the left.
---------------------------------------------------------------------------*/
void CThermometer::OnPaint() 
{
	CPaintDC dc(this); 
   CRect rectIntersect;
   CRect rectInvalid;

   // erase with gray
   rectInvalid = dc.m_ps.rcPaint;
   dc.FillRect(&rectInvalid, &m_brushLtGray);

   // Calculate the rectangles for the mercury, and draw them
   if (rectIntersect.IntersectRect(&rectInvalid, &m_rectMercury))
   {
      DrawMercury(&dc, rectInvalid);
   }
   DrawGradationLines(&dc, rectInvalid);
   if (rectIntersect.IntersectRect(&rectInvalid, &m_rectNumbers))
   {
      DrawNumbers(&dc, rectInvalid);
   }
}

/*---------------------------------------------------------------------------
DrawMercury    The current temperature, in pixels, is the top of the 
               mercury bar and the bottom of the gray bar.
               Draw them both.
               Since most of the drawing will be updates based on a change
               in temperature, be sure to get the intersection with the
               invalid rectangle.
---------------------------------------------------------------------------*/
void CThermometer::DrawMercury(CDC *pDC, CRect &rectInvalid)
{
   CRect rectIntersect;
   int inset = m_pScale->GetInset(m_rectMercury.Height());
   long nPixelsOfMercury = m_pScale->CelsiusToPixels(m_fCurrTemp);

   // Inset the mercury rectangle to match the scale
   CRect rect = m_rectMercury;
   rect.top += inset;
   rect.bottom -= inset;

   // Draw a gray rectangle above the mercury
   CRect grayRect = rect;
   grayRect.bottom = grayRect.top + nPixelsOfMercury;
   if (rectIntersect.IntersectRect(&grayRect, &rectInvalid))
   {
      pDC->SelectObject(&m_penDkGray);
	   pDC->SelectObject(&m_brushLtGray);
      pDC->Rectangle(&grayRect); 
   }

   // Draw a red rectangle to represent the mercury
   CRect redRect = rect;
   redRect.top = redRect.top + nPixelsOfMercury;
   if (rectIntersect.IntersectRect(&redRect, &rectInvalid))
   {
      pDC->SelectObject(&m_penBlack);
	   pDC->SelectObject(&m_brushMercury);
      pDC->Rectangle(&redRect); 
   }
}

/*---------------------------------------------------------------------------
DrawGradationLines Draw blue lines, alternating short and long, on either
               side of the bar of mercury. 
---------------------------------------------------------------------------*/
void CThermometer::DrawGradationLines(CDC *pDC, CRect &rectInvalid)
{
   CRect clientRect;
   GetClientRect(&clientRect);

   // The thermometer is inset within the client rectangle, in order for the
   //   scale to line up on even pixel boundaries.  Draw the thermometer
   //   in the inset rectangle.
   int nInset = m_pScale->GetInset(clientRect.Height());
   clientRect.top    += nInset;
   clientRect.bottom -= nInset;

   // Get two rectangles, for the gradation lines on the left, and on the right,
   //   
   CRect rectLeft = clientRect;
   rectLeft.right = m_rectMercury.left;
   CRect rectRight = clientRect;
   rectRight.left = m_rectMercury.right;

   // Find out if the invalid rectangles intersects either set of gradations
   BOOL  bIntersect;
   CRect rectIntersect;
   bIntersect = rectIntersect.IntersectRect(&rectInvalid, &rectLeft) ||
                rectIntersect.IntersectRect(&rectInvalid, &rectRight);

   // If there's an intersection, and if the gradation lines have a non-zero width
   //    draw a series of short horizontal blue lines on either side of 
   //    the mercury.
   if (m_gradeLong && bIntersect)
   {
	   pDC->SelectObject(&m_penLtBlue);
      int interval    = m_pScale->GetCellSize().cy;
      int halfInterval= interval / 2;
      int xLeftLong   = m_rectMercury.left  - m_gradeLong;
      int xRightLong  = m_rectMercury.right + m_gradeLong;
      int xLeftShort  = m_rectMercury.left  - m_gradeShort;
      int xRightShort = m_rectMercury.right + m_gradeShort;
      for (int y = clientRect.top + halfInterval;
           y < m_rectMercury.bottom - halfInterval;
           y += interval)
      {
         // Draw long line
         pDC->MoveTo(xLeftLong,  y);
         pDC->LineTo(m_rectMercury.left-1,  y);
         pDC->MoveTo(m_rectMercury.right+1,  y);
         pDC->LineTo(xRightLong,  y);

         // Draw short line
         int yHalf = y + halfInterval;
         pDC->MoveTo(xLeftShort,  yHalf);
         pDC->LineTo(m_rectMercury.left-1,  yHalf);
         pDC->MoveTo(m_rectMercury.right+1,  yHalf);
         pDC->LineTo(xRightShort,  yHalf);
      }
	   pDC->SelectObject(&m_penBlack);
   }
}

/*---------------------------------------------------------------------------
DrawNumbers    Draw the numeric scale.  Move from bottom to top in increments
               of cellsize (which was previously calculated with GetCellSize).
               At each position, translate the pixel position to a temperature,
               round the temperature to the nearest scale interval (modulus)
               and draw the text centered in the cell.
---------------------------------------------------------------------------*/
void CThermometer::DrawNumbers(CDC *pDC, CRect &)
{
   int oldMode = pDC->SetBkMode(TRANSPARENT);
   COLORREF oldColor = pDC->SetTextColor(RGB_GRIDLINE_COLOR);
	pDC->SelectObject(&m_font);

   CRect clientRect;
   GetClientRect(&clientRect);

   int inset       = m_pScale->GetInset(clientRect.Height());
   int nCellHeight = m_pScale->GetCellSize().cy;
   int nHalfHeight = nCellHeight / 2;
   int nDivisor    = m_pScale->GetScale();
   int yCelsius    = nHalfHeight;

   while (yCelsius < m_pScale->GetHeight())
   {
      // convert pixel position to temperature
      float fTemperature = m_pScale->PixelsToCelsius(yCelsius);
      if (m_units == FAHRENHEIT)
      {
         fTemperature = ToFahrenheit(fTemperature);
      }
      // round to nearest scale interval
      long nTemperature = CTempScale::RoundToNearestMultiple(fTemperature,
                                                             nDivisor);
      // get numeric text and center in the cell
      char text[8];
      sprintf(text, "%d", nTemperature);

      CSize size = pDC->GetTextExtent(text, strlen(text));
      int x      = m_rectNumbers.left +
                   ((m_rectNumbers.Width() - size.cx) / 2);
      int yOffset = (nCellHeight - size.cy) / 2;
      int y       = yCelsius - nHalfHeight + inset + yOffset;

      pDC->TextOut(x, y, text, strlen(text));

      // increment to next temperature and position
      yCelsius += nCellHeight;
   }
   pDC->SetBkMode(oldMode);
   pDC->SetTextColor(oldColor);
}


/*---------------------------------------------------------------------------
SetRect        Set the window rectangle
---------------------------------------------------------------------------*/
void CThermometer::SetRect(CRect &rect)
{
   // Get new client rect
   CRect clientRect(0,0,rect.Width(),rect.Height());

   // temperature numbers
   m_rectNumbers = clientRect;
   m_rectNumbers.right = m_pScale->GetCellSize().cx + 6;

   // thin rectangle enclosing bar of mercury
   int availWidth = clientRect.Width() - m_rectNumbers.Width();
   int nInset  = (2 * availWidth) / 5;
   m_rectMercury = clientRect;
   m_rectMercury.left = m_rectNumbers.right;
   m_rectMercury.DeflateRect(nInset, 0);

   // width of gradation, long and short
   m_gradeLong = (availWidth - m_rectMercury.Width()) / 2 - 2;
   m_gradeShort = m_gradeLong / 2;

   if (GetSafeHwnd() && (m_rectWindow != rect))
   {
      MoveWindow(rect, TRUE);
   }
   m_rectWindow = rect;
}   

/*---------------------------------------------------------------------------
GetDesiredSize increase given width and height as necessary to satisfy 
               minimum size requirements
---------------------------------------------------------------------------*/
void CThermometer::GetDesiredSize(int &width, int &height)
{
   HRESULT hr = S_OK;
   CSize   minSize;
                                    // text width for numeric scale
   minSize = CTempScale::GetTextCellSize();
   minSize.cx += MIN_METER_WIDTH + 6;   //   plus minimum for thermometer

   width  = max(width, minSize.cx);
   height = max(height, MIN_METER_HEIGHT);    // must have a min height
}
/*---------------------------------------------------------------------------
GetMeterWidth  get the width of the meter and the gradation lines
---------------------------------------------------------------------------*/
int CThermometer::GetMeterWidth(void)
{
   return MIN_METER_WIDTH;
}

/*---------------------------------------------------------------------------
OnTemperatureNotifier   
---------------------------------------------------------------------------*/
void CThermometer::OnTemperatureNotifier(CTemperatureNotifier *pNotifier)
{
   if ((GetSafeHwnd() != 0) && 
       (m_fCurrTemp != pNotifier->m_celsius))
   {
      CRect rectInvalid = m_rectMercury;
      long nPixels1 = m_pScale->CelsiusToPixels(m_fCurrTemp);
      long nPixels2 = m_pScale->CelsiusToPixels(pNotifier->m_celsius);
      int  nInset   = m_pScale->GetInset(m_rectMercury.Height());

      m_fCurrTemp = pNotifier->m_celsius;
      if (nPixels1 > nPixels2)
      {
         rectInvalid.top    = nPixels2-1+nInset;
         rectInvalid.bottom = nPixels1+1+nInset;
         InvalidateRect(&rectInvalid, FALSE);
      }
      else if (nPixels2 > nPixels1)
      {
         rectInvalid.top    = nPixels1-1+nInset;
         rectInvalid.bottom = nPixels2+1+nInset;
         InvalidateRect(&rectInvalid, FALSE);
      }
   }
}

/*---------------------------------------------------------------------------
OnUnitsNotifier   change from celsius to fahrenheit
---------------------------------------------------------------------------*/
void CThermometer::OnUnitsNotifier(CUnitsNotifier *pNotifier)
{
   if (m_units != pNotifier->m_units)
   {
      // change the units
      m_units = pNotifier->m_units;

      Invalidate(FALSE);
   }
}

