/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
scale.cpp   CTempScale class implementation

            Calculates the scale of temperature windows based on
            current units and window height.
---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "thermdef.h"
#include "unitsNotifier.h"
#include "requestNotifier.h"
#include "scale.h"
#include "meter.h"

#define LO_TEMP   -40     // lowest and highest temperatures
#define HI_TEMP    50     //   displayed in thermometer

/*---------------------------------------------------------------------------
CTempScale   constructor
---------------------------------------------------------------------------*/
CTempScale::CTempScale(void)
{
   m_units = CELSIUS;
   m_cellSize = CSize(0,0);
   m_loTemp = LO_TEMP;
   m_hiTemp = HI_TEMP;
}

/*---------------------------------------------------------------------------
CTempScale   destructor
---------------------------------------------------------------------------*/
CTempScale::~CTempScale(void)
{
}

/*---------------------------------------------------------------------------
CelsiusToPixels   Convert a temperature to a position on a graph or thermometer
---------------------------------------------------------------------------*/
long CTempScale::CelsiusToPixels(float fTemperature)
{
   float fPositionInRange = (fTemperature - m_loTemp) / (m_hiTemp - m_loTemp);
   long  nPixels = long(float(m_nScaleHeight) * (1.0-fPositionInRange) + 0.5);

   return nPixels;
}
/*---------------------------------------------------------------------------
PixelsToCelsius   Convert a window position to a temperature
---------------------------------------------------------------------------*/
float CTempScale::PixelsToCelsius(long nPixels)
{
   float fPositionInRange = float(1.0) - (float(nPixels) / float(m_nScaleHeight));
   float fTemperature = fPositionInRange * (m_hiTemp - m_loTemp) + m_loTemp;
   return fTemperature;
}


/*---------------------------------------------------------------------------
CalcScaleInterval   Calculate the pixels between thermometer and graph 
                    gradations, given a window height.

Side effects        This calculates a smaller window height that contains
                    an even multiple of gradations, it sets m_cellSize which is
                    the size of the gradations, and it sets m_loTemp and m_hiTemp
                    which is in even multiples of the degrees.
---------------------------------------------------------------------------*/
void CTempScale::CalcScaleInterval(int nWndHeight,       // window height
                                   THERM_UNITS units)    // temperature units
{
   // calculate minimum text cell size
   m_cellSize = GetTextCellSize();

   // get the nominal range in the current units
   m_units = units;
   float fNominalLoTemp = LO_TEMP;
   float fNominalHiTemp = HI_TEMP;
   if (m_units == FAHRENHEIT)
   {
      fNominalLoTemp = ToFahrenheit(fNominalLoTemp);
      fNominalHiTemp = ToFahrenheit(fNominalHiTemp);
   }

   // try a series of possible scales
   //   starting with 2 degrees, and incrementing 
   //   at even 5-degree intervals
   int  nIncrement = 3;
   bool bDone = false;
   m_nScale = 2; 
   while (!bDone && (m_nScale < fNominalHiTemp))
   {
      // adjust temp range to even multiples of nDegrees
      long nLoTemp = RoundToNearestMultiple(fNominalLoTemp, m_nScale);
      long nHiTemp = RoundToNearestMultiple(fNominalHiTemp, m_nScale);

      // divide temp range by m_nScale to get number of cells
      long nCells = (nHiTemp - nLoTemp) / m_nScale;
            // it should divide evenly
      ASSERT((m_nScale * nCells) == (nHiTemp - nLoTemp));

      // multiply number of cells by cell height to get min height
      long minHeight = nCells * m_cellSize.cy;

      // if window height is large enough,
      //  adjust window height to an even number of cells
      m_nScaleHeight = nWndHeight;
      if (nWndHeight >= minHeight)
      {
         // largest cell size that fits evenly in window
         m_cellSize.cy = nWndHeight / nCells;
         // then adjust window to exactly that size
         m_nScaleHeight = m_cellSize.cy * nCells;

         // shift temp range by half a scale interval
         //   so the center of the interval is on an even number
         m_loTemp = float(nLoTemp) + float(m_nScale) / 2;
         m_hiTemp = float(nHiTemp) + float(m_nScale) / 2;

         // all temperatures are kept in celsius until displayed
         //   so convert back if necessary
         if (m_units == FAHRENHEIT)
         {
            m_loTemp = ToCelsius(m_loTemp);
            m_hiTemp = ToCelsius(m_hiTemp);
         }
         bDone = true;
      }
      // try 2, then 5, then 10, then 15, etc.
      if (!bDone)
      {
         if (m_nScale == 2)
         {
            m_nScale = 5;
         }
         else
         {
            m_nScale += 5;
         }
      }
   }
   ASSERT(bDone);
}

/*---------------------------------------------------------------------------
GetTextCellSize   Calculate min cell size in the standard font
---------------------------------------------------------------------------*/
CSize CTempScale::GetTextCellSize(void)
{
   // calculate minimum height of the cell 
   //   given the font for the numeric scale
   CDC     dc;	
   CFont   font;
   CSize   cellSize(24,24);
	if (dc.CreateIC("DISPLAY", 0,0,0))
   {
      font.CreateStockObject(ANSI_VAR_FONT);
      dc.SelectObject(&font);
      cellSize = dc.GetTextExtent("999", 3);
      // make sure it's an even number
      //cellSize.cy += (cellSize.cy & 0x01) ? 1 : 0;  
   }
   return cellSize;
}

/*---------------------------------------------------------------------------
RoundToNearestMultiple  
---------------------------------------------------------------------------*/
long CTempScale::RoundToNearestMultiple(float fInput, int nDivisor)
{
   // round to nearest scale interval
   long nOutput = Round(fInput);
   long nError = nOutput % nDivisor;
   if (nError < (nDivisor / 2))
   {
      nOutput -= nError;
   }
   else
   {
      nOutput += nDivisor - nError;
   }
   return nOutput;
}

