/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
slider.cpp  CSetPointSlider class definition

This overrides the Windows common slider control to move the setpoint
up and down.
---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "thermdef.h"
#include "slider.h"
#include "temperatureNotifier.h"
#include "unitsNotifier.h"
#include "requestNotifier.h"
#include "setpointNotifier.h"
#include "scale.h"
#include "simerr_notify.h"       // error simulation

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*---------------------------------------------------------------------------
Message Map
---------------------------------------------------------------------------*/
BEGIN_MESSAGE_MAP(CSetPointSlider, CSliderCtrl)
	//{{AFX_MSG_MAP(CSetPointSlider)
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	ON_WM_KEYUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*---------------------------------------------------------------------------
CSetPointSlider constructor
---------------------------------------------------------------------------*/
CSetPointSlider::CSetPointSlider(void)
                :CSliderCtrl(), CThermSubscriber()
{
   m_rectWindow.SetRectEmpty();

   m_units = CELSIUS;   
   m_fSetPoint = ToCelsius(75.0);
   m_bButtonDown = false;
   m_pScale = GetAppScale();
}

/*---------------------------------------------------------------------------
CSetPointSlider destructor
---------------------------------------------------------------------------*/
CSetPointSlider::~CSetPointSlider(void)
{
}

/*---------------------------------------------------------------------------
Init           create and size the window and 
               create pens & brushes for drawing
---------------------------------------------------------------------------*/
HRESULT CSetPointSlider::Init(CWnd *pParent, UINT nID)
{
   HRESULT hr = S_OK;
   BOOL bCreated = FALSE;
   
   if (SUCCEEDED(hr))
   {
      // create the window
      bCreated = Create(TBS_VERT | TBS_NOTICKS | 
                        WS_CHILD | WS_VISIBLE,
                        m_rectWindow, pParent, nID);
      bCreated = SimErrCreateWin(SIMLINE, bCreated, this);
      if (!bCreated)
      {
         hr = E_OUTOFMEMORY;
      }
      else
      {
         SetTemperatureRange();

         // find out the temperature
         Subscribe(UNITS_NOTIFIER);
         Subscribe(REQUEST_NOTIFIER);
         RequestNotifierPost(UNITS_NOTIFIER);
      }
   }

   return hr;
}

/*---------------------------------------------------------------------------
SetRect        Set the window rectangle
---------------------------------------------------------------------------*/
void CSetPointSlider::SetRect(CRect &rect)
{
   if (GetSafeHwnd() && (m_rectWindow != rect))
   {
      MoveWindow(rect, TRUE);
   }
   m_rectWindow = rect;
   SetTemperatureRange();
}   

/*---------------------------------------------------------------------------
GetDesiredSize increase given width and height as necessary to satisfy 
               minimum size requirements
---------------------------------------------------------------------------*/
void CSetPointSlider::GetDesiredSize(int &width, int &height)
{
   HRESULT hr = S_OK;
   CSize   minSize;
                                    // text width for numeric scale
   minSize.cx = MIN_SLIDER_WIDTH;   //   plus minimum for thermometer

   width  = max(width, minSize.cx);
   height = max(height, MIN_METER_HEIGHT);    // must have a min height
}

/*---------------------------------------------------------------------------
OnUnitsNotifier   change from celsius to fahrenheit
---------------------------------------------------------------------------*/
void CSetPointSlider::OnUnitsNotifier(CUnitsNotifier *pNotifier)
{
   if (m_units != pNotifier->m_units)
   {
      // change the units
      m_units = pNotifier->m_units;

      SetTemperatureRange();
   }
}

/*---------------------------------------------------------------------------
SetTemperatureRange   Set the range of the slider to the temperature range
---------------------------------------------------------------------------*/
void CSetPointSlider::SetTemperatureRange(void)
{
   if (m_hWnd)
   {
      CRect rectChannel;
      CRect rectThumb;
      int   thumbLength;
         
      thumbLength   = ::SendMessage(m_hWnd, TBM_GETTHUMBLENGTH, 0, 0);
      GetThumbRect(&rectThumb);
      GetChannelRect(&rectChannel);
      thumbLength = rectThumb.Height();

      int nScaleHeight  = m_pScale->GetHeight();
      int nSliderHeight = rectChannel.Width();
      int nSliderInset  = nScaleHeight - nSliderHeight;
      int top           = nSliderInset;
      int bottom        = nScaleHeight - nSliderInset;
      m_fHiTemp = m_pScale->PixelsToCelsius(top);
      m_fLoTemp = m_pScale->PixelsToCelsius(bottom);

      // To ensure that the slider increments in the 
      //   degrees of the current temperature units
      //   the slider range is in the current units
      if (m_units == FAHRENHEIT)
      {
         m_fHiTemp = ToFahrenheit(m_fHiTemp);
         m_fLoTemp = ToFahrenheit(m_fLoTemp);
         SetRange(0, Round(m_fHiTemp - m_fLoTemp));

         // setpoint is always in celsius
         SetPos(Round(m_fHiTemp - ToFahrenheit(m_fSetPoint)));
      }
      else
      {
         SetRange(0, Round(m_fHiTemp - m_fLoTemp));
         SetPos(Round(m_fHiTemp - m_fSetPoint));
      }
   }
}

/*---------------------------------------------------------------------------
GetNewSetPoint    
---------------------------------------------------------------------------*/
float CSetPointSlider::GetNewSetPoint(void)
{
   float fPos = (float) GetPos();
   float fNewSetPoint = m_fHiTemp - fPos;
   if (m_units == FAHRENHEIT)
   {
      fNewSetPoint = ToCelsius(fNewSetPoint);
   }
   return fNewSetPoint;
}


/*---------------------------------------------------------------------------
OnRequestNotifier    someone wants the current thermostat set point
---------------------------------------------------------------------------*/
void CSetPointSlider::OnRequestNotifier(CRequestNotifier *pNotifier)
{
   if (pNotifier->m_nClass == SETPOINT_NOTIFIER)
   {
      if (pNotifier->m_bSend)
      {
         SendSetPointNotifier(m_fSetPoint);
      }
      else
      {
         PostSetPointNotifier(m_fSetPoint);
      }
   }
}


/*---------------------------------------------------------------------------
OnLButtonDown    
---------------------------------------------------------------------------*/
void CSetPointSlider::OnLButtonDown(UINT nFlags, CPoint point) 
{
	CSliderCtrl::OnLButtonDown(nFlags, point);
   m_bButtonDown = true;
}

/*---------------------------------------------------------------------------
OnLButtonUp    
---------------------------------------------------------------------------*/
void CSetPointSlider::OnLButtonUp(UINT nFlags, CPoint point) 
{
   m_bButtonDown = false;
	CSliderCtrl::OnLButtonUp(nFlags, point);
   float fNewSetPoint = GetNewSetPoint();
   if (fNewSetPoint != m_fSetPoint)
   {
      m_fSetPoint = fNewSetPoint;
      PostSetPointNotifier(m_fSetPoint);
   }
}

/*---------------------------------------------------------------------------
OnMouseMove    
---------------------------------------------------------------------------*/
void CSetPointSlider::OnMouseMove(UINT nFlags, CPoint point) 
{
	CSliderCtrl::OnMouseMove(nFlags, point);
   // if button is down, report any changes
   if (m_bButtonDown)
   {
      float fNewSetPoint = GetNewSetPoint();
      // only send the big changes while we're moving around
      if (Round(fNewSetPoint) != Round(m_fSetPoint))
      {
         m_fSetPoint = fNewSetPoint;
         SendSetPointNotifier(m_fSetPoint);
      }
   }
}

/*---------------------------------------------------------------------------
OnKeyUp    
---------------------------------------------------------------------------*/
void CSetPointSlider::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	CSliderCtrl::OnKeyUp(nChar, nRepCnt, nFlags);
   float fNewSetPoint = GetNewSetPoint();
   if (fNewSetPoint != m_fSetPoint)
   {
      m_fSetPoint = fNewSetPoint;
      PostSetPointNotifier(m_fSetPoint);
   }
}
