/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
temperature.cpp   CTemperature class implementation

This class simulates a real-time sensor for temperature data.  This is
designed to run in its own  thread so access to real-time data is not held 
up by the user-interface.
---------------------------------------------------------------------------*/
#include <stdlib.h>
#include "stdafx.h"
#include "thermdef.h"
#include "temperatureNotifier.h"
#include "setpointNotifier.h"
#include "requestNotifier.h"
#include "temperature.h"
#include "simerr_notify.h"       // error simulation

static CTemperature *g_pTemperature; // there is ONE copy of the temperature device
const double TEMP_RANGE = HI_TEMP - LO_TEMP;
const double HALF_RANGE = TEMP_RANGE / 2;

                  //====================
                  // Static Thread Proc
                  //====================

/*---------------------------------------------------------------------------
TemperatureThreadProc  Thread proc for running the temperature sensor.
---------------------------------------------------------------------------*/
UINT TemperatureThreadProc(LPVOID argEvent)
{
   CEvent *pEvent = (CEvent *) argEvent;
   CTemperature *pTemperature = new CTemperature();
   pTemperature = SimErrNew(SIMLINE, pTemperature);
   if (pTemperature == 0)
   {
      AfxGetApp()->PostThreadMessage(TEMPERATURE_THREAD, THREAD_FAILED, 0);  
      return 1;
   }
   bool bContinue = true;
   while (bContinue)
   {
      switch(::WaitForSingleObject(*pEvent, 50))
      {
      case WAIT_OBJECT_0:
         // Main app is telling us to shut down
         bContinue = false;
         break;

      case WAIT_TIMEOUT:
         pTemperature->PollSensor(GetTickCount());
         break;
      }
   }
   return 0;
}

                  //=============================
                  // Construction and Assignment
                  //=============================

//---------------------------------------------------------------------------
// Void constructor
CTemperature::CTemperature(void)
             :CThermSubscriber()
{
   m_fTemperature = 34.0;
   m_fSetPoint = float(LO_TEMP + HALF_RANGE);
   InitializeCriticalSection(&m_csTemperature);
   InitializeCriticalSection(&m_csSetPoint);
   Subscribe(REQUEST_NOTIFIER);
   Subscribe(SETPOINT_NOTIFIER);
   RequestNotifierPost(SETPOINT_NOTIFIER);
}
//---------------------------------------------------------------------------
// Destructor
CTemperature::~CTemperature(void)
{
   DeleteCriticalSection(&m_csTemperature);
   DeleteCriticalSection(&m_csSetPoint);
}

                  //========================
                  // Notification Responses
                  //========================

/*---------------------------------------------------------------------------
PollSensor  Poll the temperature sensor and notify subscribers if the 
            temperature has changed.
            (of course we fake all that with random numbers)
---------------------------------------------------------------------------*/
void CTemperature::PollSensor(DWORD dwTime)
{

#ifdef RELEASE_3
   EnterCriticalSection(&m_csSetPoint);
   double dMidTemp = m_fSetPoint;
   LeaveCriticalSection(&m_csSetPoint);
#else
   const double dMidTemp = LO_TEMP + HALF_RANGE;
#endif

   static DWORD g_lastTemperatureChange = 0;

   if ((dwTime - g_lastTemperatureChange) > 100)
   {  // pick a small random change
      double dChange = ((rand() % 40) - 20.0) / 25.0;
      double dDistance;
      double dAbsDistance;

#ifdef RELEASE_3
      // The temperature moves to the middle of the temerature range
      // For the setpoint release, if we're far from the setpoint, move
      //   the midpoint farther to encourage a rapid return to the setpoint.
      dDistance = m_fTemperature - dMidTemp;
      if (dDistance > 2)
      {  // too hot, turn off the furnace (pull the temperature down)
         dMidTemp -= 10;
      }
      else if (dDistance < 2)
      {  // too cold, turn on the furnace
         dMidTemp += 10;
      }
#endif

      // As the temperature moves away from the middle of the range,
      //   decrease any changes that would increase that distance.
      dDistance = m_fTemperature - dMidTemp;
      if ((dChange * dDistance) > 0)
      {  // when dChange and dDistance have the same arithmetic sign
         //   diminish dChange in proportion to its distance from the mid-temp
         dAbsDistance = (dDistance < 0) ? -dDistance : dDistance;
         double dFactor = (HALF_RANGE - dAbsDistance)  / HALF_RANGE;
         dChange = dChange * dFactor;
      }

      // Now change the temperature
      EnterCriticalSection(&m_csTemperature);
      m_fTemperature += (float) dChange;
      float fTemperature = m_fTemperature;
      LeaveCriticalSection(&m_csTemperature);

      PostTemperatureNotifier(fTemperature);
      g_lastTemperatureChange = dwTime;
   }
}

/*---------------------------------------------------------------------------
OnRequestNotifier    someone wants a notifier with the current temperature
---------------------------------------------------------------------------*/
void CTemperature::OnRequestNotifier(CRequestNotifier *pNotifier)
{
   if (pNotifier->m_nClass == TEMPERATURE_NOTIFIER)
   {
      EnterCriticalSection(&m_csTemperature);
      float fTemperature = m_fTemperature;
      LeaveCriticalSection(&m_csTemperature);

      if (pNotifier->m_bSend)
      {
         SendTemperatureNotifier(fTemperature);
      }
      else
      {
         PostTemperatureNotifier(fTemperature);
      }
   }
}

/*---------------------------------------------------------------------------
OnSetPointNotifier   thermostat setpoint has changed
---------------------------------------------------------------------------*/
void CTemperature::OnSetPointNotifier(CSetPointNotifier *pNotifier)
{
   EnterCriticalSection(&m_csSetPoint);
   m_fSetPoint = pNotifier->m_celsius;
   LeaveCriticalSection(&m_csSetPoint);
}
