/*---------------------------------------------------------------------------
         Copyright 1997    On Target Software    All Rights Reserved

  Dave Pomerantz, Nov 1997.  Updates available at http://www.targetsoft.com
-----------------------------------------------------------------------------
ThermView.cpp   CThermView class implementation

Create and position all the display objects:

RELEASE 1
   thermometer    bar of mercury, numeric scale and gradation lines
   text degrees   text field with temperature degrees in current units
   units button   switch units between Celsius and Fahrenheit

RELEASE 2
   graph          graph of temperature over time

RELEASE 3
   slider         control that adjusts setpoint
   text setpoint  text field with setpoint degrees in current units
---------------------------------------------------------------------------*/
#include "stdafx.h"
#include "thermostat.h"
#include "doc.h"
#include "view.h"

#include "temperature.h"
#include "meter.h"
#include "textDegrees.h"
#include "btnUnits.h"
#include "scale.h"
#include "unitsNotifier.h"
#include "requestNotifier.h"
#include "simerr_notify.h"       // error simulation

#ifdef RELEASE_2
#include "graph.h"
#endif

#ifdef RELEASE_3
#include "slider.h"
#include "textSetPoint.h"
#endif


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*---------------------------------------------------------------------------
Message Map
---------------------------------------------------------------------------*/
IMPLEMENT_DYNCREATE(CThermView, CView)

BEGIN_MESSAGE_MAP(CThermView, CView)
	//{{AFX_MSG_MAP(CThermView)
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/*---------------------------------------------------------------------------
CThermView   constructor

High priority subscriber.  This object receives notifications first.
See the OnUnitsNotifier for an explanation.
---------------------------------------------------------------------------*/
CThermView::CThermView()
           :CThermSubscriber(PRIORITY_HIGH)
{
   m_bInitialized = false;
   m_units = CELSIUS;
   m_pTemperatureThread = 0;
   m_pThermometer = 0;
   m_pTxtDegrees  = 0;
   m_pBtnUnits    = 0;
#ifdef RELEASE_2
   m_pGraph       = 0;
#endif
#ifdef RELEASE_3
   m_pSlider      = 0;
   m_pTextSetPoint = 0;
#endif

   // Calculate initial size of control
   int width = 0;
   int height = 0;
}

/*---------------------------------------------------------------------------
CThermView destructor
---------------------------------------------------------------------------*/
CThermView::~CThermView()
{
   if (m_pTemperatureThread)
   {  // terminate temperature sensor thread
      m_hEvent.SetEvent(); // signal the temperature thread to exit
      ::WaitForSingleObject(m_pTemperatureThread->m_hThread, INFINITE);
      delete m_pTemperatureThread;
   }
   if (m_pThermometer)
   {
      delete m_pThermometer;
   }
   if (m_pTxtDegrees)
   {
      delete m_pTxtDegrees;
   }
   if (m_pBtnUnits)
   {
      delete m_pBtnUnits;
   }
#ifdef RELEASE_2
   if (m_pGraph)
   {
      delete m_pGraph;
   }
#endif
#ifdef RELEASE_3
   if (m_pSlider)
   {
      delete m_pSlider;
   }
   if (m_pTextSetPoint)
   {
      delete m_pTextSetPoint;
   }
#endif
}

/*---------------------------------------------------------------------------
Init           create sensor object and display objects
---------------------------------------------------------------------------*/
HRESULT CThermView::Init(void)
{
   HRESULT hr = S_OK;
   static bool bInitializing = false;
   if (!m_bInitialized && !bInitializing)
   {
      bInitializing = true;
      hr = InitSensor();
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
      if (SUCCEEDED(hr))
      {
         hr = InitDisplay();
         hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
      }
      if (SUCCEEDED(hr))
      {
         m_bInitialized = true;
      }
   }
   return hr;
}

/*---------------------------------------------------------------------------
InitSensor     create temperature sensor running in its own thread
---------------------------------------------------------------------------*/
HRESULT CThermView::InitSensor(void)
{
   HRESULT hr = S_OK;
   extern UINT TemperatureThreadProc(LPVOID);

   if (m_pTemperatureThread == 0)
   {
      m_pTemperatureThread = new CWinThread(TemperatureThreadProc, &m_hEvent);
      m_pTemperatureThread = SimErrNew(SIMLINE, m_pTemperatureThread);
      if (m_pTemperatureThread)
      {
         m_pTemperatureThread->m_bAutoDelete = FALSE;
         BOOL bOK = m_pTemperatureThread->CreateThread();
              bOK = SimErrZero(SIMLINE, bOK);
         if (!bOK)
         {
            hr = E_OUTOFMEMORY;
         }
      }
      else
      {
         hr = E_OUTOFMEMORY;
      }
   }
   return hr;
}

/*---------------------------------------------------------------------------
InitDisplay    create temperature display objects
---------------------------------------------------------------------------*/
HRESULT CThermView::InitDisplay(void)
{
   HRESULT hr = E_OUTOFMEMORY;
   int nWndID = FIRST_WINDOW_ID;

   // Create and initialize all Release 1 objects
   if (m_pThermometer == 0)
   {
      m_pThermometer = new CThermometer();
      m_pThermometer = SimErrNew(SIMLINE, m_pThermometer);
   }
   if (m_pTxtDegrees == 0)
   {
      m_pTxtDegrees = new CTextDegrees();
      m_pTxtDegrees = SimErrNew(SIMLINE, m_pTxtDegrees);
   }
   if (m_pBtnUnits == 0)
   {
      m_pBtnUnits = new CBtnUnits();
      m_pBtnUnits = SimErrNew(SIMLINE, m_pBtnUnits);
   }
   if (m_pThermometer &&
       m_pTxtDegrees  &&
       m_pBtnUnits)
   {
      hr = m_pTxtDegrees->Init(this, nWndID++);
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
   }
   if (SUCCEEDED(hr))
   {
      hr = m_pThermometer->Init(this, nWndID++);
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
   }
   if (SUCCEEDED(hr))
   {
      hr = m_pBtnUnits->Init(this, nWndID++);
      hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
   }

#ifdef RELEASE_2
   // Create and initialize all Release 2 objects
   if (SUCCEEDED(hr))
   {
      if (m_pGraph==0)
      {
         m_pGraph = new CGraph();
         m_pGraph = SimErrNew(SIMLINE, m_pGraph);
      }
      if (m_pGraph)
      {
         hr = m_pGraph->Init(this, nWndID++);
         hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
      }
      else
      {
         hr = E_OUTOFMEMORY;
      }
   }
#endif
#ifdef RELEASE_3
   // Create and initialize all Release 3 objects
   if (SUCCEEDED(hr))
   {
      if (m_pSlider==0)
      {
         m_pSlider = new CSetPointSlider();
         m_pSlider = SimErrNew(SIMLINE, m_pSlider);
      }
      if (m_pSlider)
      {
         hr = m_pSlider->Init(this, nWndID++);
         hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
      }
      else
      {
         hr = E_OUTOFMEMORY;
      }
   }
   if (SUCCEEDED(hr))
   {
      if (m_pTextSetPoint==0)
      {
         m_pTextSetPoint = new CTextSetPoint();
         m_pTextSetPoint = SimErrNew(SIMLINE, m_pTextSetPoint);
      }
      if (m_pTextSetPoint)
      {
         hr = m_pTextSetPoint->Init(this, nWndID++);
         hr = SimErrResult(SIMLINE, hr, E_OUTOFMEMORY);
      }
      else
      {
         hr = E_OUTOFMEMORY;
      }
   }
#endif
   // After all objects are successfully created, size and position them
   if (SUCCEEDED(hr))
   {
      CRect client;
      GetClientRect(&client);
      int width = client.Width();
      int height = client.Height();
      Resize(width, height);
#ifdef RELEASE_3
      // To make the slider control visible, give it the focus.
      if (m_pSlider && m_pSlider->GetSafeHwnd())
      {
         m_pSlider->SetFocus();
      }
#endif
   }

   return hr;
}

/*---------------------------------------------------------------------------
GetSize         get minimum size required by thermostat
---------------------------------------------------------------------------*/
void CThermView::GetSize(int &width, int &height) 
{
   // determine the sizes needed by the two buttons
   //   this calculates min width and we add to the height
   int degreesHeight = 0;
   int unitsHeight   = 0;
   int thermHeight   = 0;
   int thermWidth    = 0;
   int sliderWidth   = 0;
   int origWidth     = width;
   int meterWidth    = CThermometer::GetMeterWidth();

   CTextDegrees::GetDesiredSize(thermWidth, degreesHeight);
   CBtnUnits::GetDesiredSize(thermWidth, unitsHeight);

   thermHeight = height - (degreesHeight + unitsHeight);
   CThermometer::GetDesiredSize(thermWidth, thermHeight);
#ifdef RELEASE_3
   int dummy;
   CSetPointSlider::GetDesiredSize(sliderWidth, thermHeight);
   CTextSetPoint::GetDesiredSize(sliderWidth, dummy);
   sliderWidth = meterWidth;
#endif

   width = thermWidth + sliderWidth;
   height = thermHeight + degreesHeight + unitsHeight + 4;
}

/*---------------------------------------------------------------------------
Resize         size and position temperature display objects
---------------------------------------------------------------------------*/
void CThermView::Resize(int &width, int &height) 
{
   int degreesHeight = 0;  // height of text degrees window
   int unitsHeight   = 0;  // height of Celsius button
   int thermHeight   = 0;  // height and width
   int thermWidth    = 0;  //   of the thermometer itself
   int sliderWidth   = 0;  // width of the setpoint slider
   int origWidth     = width;

                           // width of the mercury bar and gradations
                           //   without the numeric scale, so we can
                           //   center the button over just that part.
   int meterWidth    = CThermometer::GetMeterWidth();

   // This is pretty sloppy and needs to be rewritten.
   CTextDegrees::GetDesiredSize(thermWidth, degreesHeight);
   CBtnUnits::GetDesiredSize(thermWidth, unitsHeight);

         // thermometer fits between text degrees and Celsius button
   thermHeight = height - (degreesHeight + unitsHeight);
   CThermometer::GetDesiredSize(thermWidth, thermHeight);
#ifdef RELEASE_3
   int dummy;
   CSetPointSlider::GetDesiredSize(sliderWidth, thermHeight);
   CTextSetPoint::GetDesiredSize(sliderWidth, dummy);
   sliderWidth = meterWidth;
#endif

   width = thermWidth + sliderWidth;
   height = thermHeight + degreesHeight + unitsHeight + 4;

   m_thermRect.SetRect(0,0,thermWidth + sliderWidth,height);
   GetAppScale()->CalcScaleInterval(thermHeight, m_units);

   CRect objRect = m_thermRect;
#ifdef RELEASE_2
   // if displaying graph, move thermometer to the right
   if (origWidth > width)
   {
      objRect.OffsetRect(origWidth - width, 0);
   }
#endif
   objRect.DeflateRect(2,2);
   CRect saveRect = objRect;

   // position degrees display on top
   objRect.bottom = objRect.top + degreesHeight-1;
   objRect.left   = objRect.right - meterWidth;
   if (m_pTxtDegrees)
   {  
      m_pTxtDegrees->SetRect(objRect);
   }

   // position thermometer
   objRect = saveRect;
   objRect.top    += degreesHeight-1;
   objRect.bottom  = objRect.top + thermHeight-1;
   objRect.left    = objRect.right - thermWidth;
   if (m_pThermometer)
   {  
      m_pThermometer->SetRect(objRect);
   }

#ifdef RELEASE_3
   // position set point text
   objRect = saveRect;
   objRect.bottom = objRect.top + degreesHeight-1;
   objRect.right  = objRect.left + sliderWidth;
   objRect.left   += 2;
   if (m_pTextSetPoint)
   {  
      m_pTextSetPoint->SetRect(objRect);
   }
   // position set point slider
   objRect = saveRect;
   objRect.top    += degreesHeight-1;
   objRect.bottom  = objRect.top + thermHeight-1;
   objRect.right   = objRect.left + sliderWidth;
   objRect.left   += 2;
   if (m_pSlider)
   {  
      m_pSlider->SetRect(objRect);
   }
#endif

   // position units button on bottom
   objRect = saveRect;
   objRect.top  += degreesHeight + thermHeight;
   objRect.bottom -= 2;
   objRect.left  = objRect.right - meterWidth;
   if (m_pBtnUnits)
   {
      m_pBtnUnits->SetRect(objRect);
   }

#ifdef RELEASE_2
   // position temperature graph
   objRect = m_thermRect;
   objRect.right = origWidth - width;
   objRect.DeflateRect(2,2);
   objRect.top    += degreesHeight-1;
   objRect.bottom  = objRect.top + thermHeight-1;
   if (m_pGraph)
   {
      m_pGraph->SetRect(objRect);
   }
#endif
   if (origWidth > width)
   {
      m_thermRect.OffsetRect(origWidth - width, 0);
   }

}

/*---------------------------------------------------------------------------
OnDraw         paint gray behind object
---------------------------------------------------------------------------*/
void CThermView::OnDraw(CDC* pDC)
{
   CRect  rectClient;
   CBrush grayBrush;

   GetClientRect(&rectClient);
   grayBrush.CreateStockObject(LTGRAY_BRUSH);
   pDC->FillRect(&rectClient, &grayBrush);

   CView::OnDraw(pDC);
}

/*---------------------------------------------------------------------------
AssertValid    debug only
---------------------------------------------------------------------------*/
#ifdef _DEBUG
void CThermView::AssertValid() const
{
	CView::AssertValid();
}

/*---------------------------------------------------------------------------
Dump    debug only
---------------------------------------------------------------------------*/
void CThermView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

/*---------------------------------------------------------------------------
GetDocument    release uses inline function
---------------------------------------------------------------------------*/
CDoc* CThermView::GetDocument() 
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CDoc)));
	return (CDoc*)m_pDocument;
}
#endif //_DEBUG

/*---------------------------------------------------------------------------
OnSize    resize and reposition child windows
---------------------------------------------------------------------------*/
void CThermView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);
	
	Resize(cx, cy);
}

/*---------------------------------------------------------------------------
OnInitialUpdate         Create child windows
---------------------------------------------------------------------------*/
void CThermView::OnInitialUpdate() 
{
   Subscribe(UNITS_NOTIFIER);
   if (SUCCEEDED(Init()))
   {
      RequestNotifier(UNITS_NOTIFIER);
	   CView::OnInitialUpdate();
   }
   else
   {
      AfxMessageBox("Unable to start Thermostat application");
      AfxGetMainWnd()->PostMessage(WM_CLOSE, 0, 0);
   }
}

/*---------------------------------------------------------------------------
OnUnitsNotifier   change from celsius to fahrenheit
---------------------------------------------------------------------------*/
void CThermView::OnUnitsNotifier(CUnitsNotifier *pNotifier)
{
   if (m_units != pNotifier->m_units)
   {
      // change the units
      m_units = pNotifier->m_units;
      if (m_pThermometer && (m_pThermometer->GetSafeHwnd() != 0))
      {
         CRect thermRect;
         m_pThermometer->GetClientRect(&thermRect);
         GetAppScale()->CalcScaleInterval(thermRect.Height(), m_units);
      }
   }
}
